@@ -20,7 +20,6 @@
 from .cases import tokamak, torpex
 from .core.equilibrium import Point2D, EquilibriumRegion, Equilibrium, SolutionError
 from .core.mesh import MultiLocationArray, MeshRegion, Mesh, BoutMesh
-from .utils.hypnotoad_options import HypnotoadOptions, HypnotoadInternalOptions
 from .__version__ import get_versions
 
 __version__ = get_versions()["version"]
@@ -36,7 +35,5 @@ __all__ = [
     "MeshRegion",
     "Mesh",
     "BoutMesh",
-    "HypnotoadOptions",
-    "HypnotoadInternalOptions",
     "__version__",
 ]
